from http import HTTPMethod, HTTPStatus
from typing import Type

import pydantic
from asgikit.requests import Request, read_form, read_json
from pydantic import BaseModel as PydanticModel

from selva.web.converter.decorator import register_from_request
from selva.web.exception import HTTPBadRequestException, HTTPException


@register_from_request(PydanticModel)
class PydanticModelFromRequest:
    async def from_request(
        self,
        request: Request,
        original_type: Type[PydanticModel],
        _parameter_name,
        _metadata=None,
    ) -> PydanticModel:
        if request.method not in (HTTPMethod.POST, HTTPMethod.PUT, HTTPMethod.PATCH):
            raise TypeError(
                "Pydantic model parameter on method that does not accept body"
            )

        # TODO: make request body decoding extensible
        if "application/json" in request.content_type:
            data = await read_json(request)
        elif "application/x-www-form-urlencoded" in request.content_type:
            data = await read_form(request)
        else:
            raise HTTPException(status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE)

        try:
            return original_type.model_validate(data)
        except pydantic.ValidationError as err:
            raise HTTPBadRequestException() from err


@register_from_request(list[PydanticModel])
class PydanticModelListFromRequest:
    async def from_request(
        self,
        request: Request,
        original_type: Type[list[PydanticModel]],
        _parameter_name,
        _metadata=None,
    ) -> list[PydanticModel]:
        if request.method not in (HTTPMethod.POST, HTTPMethod.PUT, HTTPMethod.PATCH):
            raise TypeError("Pydantic parameter on method that does not accept body")

        if "application/json" in request.content_type:
            data = await read_json(request)
        else:
            raise HTTPException(status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE)

        adapter = pydantic.TypeAdapter(original_type)

        try:
            return adapter.validate_python(data)
        except pydantic.ValidationError as err:
            raise HTTPBadRequestException() from err



from http import HTTPStatus


class HTTPException(Exception):
    status: HTTPStatus

    def __init__(
        self,
        *,
        status: HTTPStatus = None,
        headers: dict[str, str] = None,
    ):
        if status:
            self.status = status

        self.headers = headers or {}


class WebSocketException(Exception):
    def __init__(self, code: int, reason: str = None) -> None:
        self.code = code
        self.reason = reason or ""


class HTTPBadRequestException(HTTPException):
    status = HTTPStatus.BAD_REQUEST


class HTTPNotFoundException(HTTPException):
    status = HTTPStatus.NOT_FOUND


class HTTPUnauthorizedException(HTTPException):
    status = HTTPStatus.UNAUTHORIZED


class HTTPForbiddenException(HTTPException):
    status = HTTPStatus.FORBIDDEN


class HTTPInternalServerException(HTTPException):
    status = HTTPStatus.INTERNAL_SERVER_ERROR



import inspect
from collections import OrderedDict
from http import HTTPMethod

from loguru import logger

from selva.web.exception import HTTPNotFoundException
from selva.web.routing.decorator import (
    ACTION_ATTRIBUTE,
    CONTROLLER_ATTRIBUTE,
    ActionInfo,
    ControllerInfo,
)
from selva.web.routing.exception import (
    ControllerWithoutDecoratorError,
    DuplicateRouteError,
)
from selva.web.routing.route import Route, RouteMatch


def _path_with_prefix(path: str, prefix: str):
    path = path.strip("/")
    prefix = prefix.strip("/")
    return f"{prefix}/{path}"


class Router:
    def __init__(self):
        self.routes: OrderedDict[str, Route] = OrderedDict()

    def route(self, controller: type):
        controller_info: ControllerInfo = getattr(
            controller, CONTROLLER_ATTRIBUTE, None
        )

        if not controller_info:
            raise ControllerWithoutDecoratorError(controller)

        path_prefix = controller_info.path

        logger.trace(
            "controller registered at {}: {}.{}",
            path_prefix or "/",
            controller.__module__,
            controller.__qualname__,
        )

        for name, action in inspect.getmembers(controller, inspect.isfunction):
            action_info: ActionInfo = getattr(action, ACTION_ATTRIBUTE, None)
            if not action_info:
                continue

            action_type = action_info.type
            method: HTTPMethod = action_type.value  # type: ignore
            path = action_info.path
            path = _path_with_prefix(path, path_prefix)
            route_name = f"{controller.__module__}.{controller.__qualname__}:{name}"

            route = Route(method, path, controller, action, route_name)

            for current_route in self.routes.values():
                if (
                    current_route.action != route.action
                    and current_route.method == route.method
                    and current_route.regex == route.regex
                ):
                    raise DuplicateRouteError(route.name, current_route.name)

            self.routes[route_name] = route
            logger.trace(
                "action '{}.{}:{}' registered at '{} {}'",
                controller.__module__,
                controller.__qualname__,
                route.action.__name__,
                route.method,
                route.path,
            )

    def match(self, method: HTTPMethod | None, path: str) -> RouteMatch | None:
        for route in self.routes.values():
            if (match := route.match(method, path)) is not None:
                return RouteMatch(route, method, path, match)

        return None

    def reverse(self, name: str, **kwargs) -> str:
        if route := self.routes.get(name):
            return route.reverse(**kwargs)

        raise HTTPNotFoundException()



class DuplicateRouteError(Exception):
    def __init__(self, route1: str, route2: str):
        super().__init__(f"path for {route1} clashes with {route2}")


class ControllerWithoutDecoratorError(Exception):
    def __init__(self, controller: type):
        super().__init__(
            f"{controller.__module__}.{controller.__qualname__}"
            " is not decorated with @controller"
        )



import functools
import inspect
import typing
from http import HTTPStatus
from typing import Any
from uuid import uuid4

from asgikit.errors.websocket import WebSocketDisconnectError, WebSocketError
from asgikit.requests import Request
from asgikit.responses import respond_status, respond_text
from asgikit.websockets import WebSocket
from loguru import logger

from selva._util.base_types import get_base_types
from selva._util.import_item import import_item
from selva._util.maybe_async import maybe_async
from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.web.converter import (
    from_request_impl,
    param_converter_impl,
    param_extractor_impl,
)
from selva.web.converter.error import (
    MissingFromRequestImplError,
    MissingParamConverterImplError,
    MissingRequestParamExtractorImplError,
)
from selva.web.converter.from_request import FromRequest
from selva.web.converter.param_converter import ParamConverter
from selva.web.converter.param_extractor import ParamExtractor
from selva.web.exception import HTTPException, HTTPNotFoundException, WebSocketException
from selva.web.exception_handler import ExceptionHandler
from selva.web.middleware import Middleware
from selva.web.routing.decorator import CONTROLLER_ATTRIBUTE
from selva.web.routing.router import Router


def _is_controller(arg) -> bool:
    return inspect.isclass(arg) and hasattr(arg, CONTROLLER_ATTRIBUTE)


def _is_module(arg) -> bool:
    return inspect.ismodule(arg) or isinstance(arg, str)


class Selva:
    """Entrypoint class for a Selva Application

    Will try to automatically import and register a module called "application".
    Other modules and classes can be registered using the "register" method
    """

    def __init__(self, settings: Settings):
        self.di = Container()
        self.di.define(Container, self.di)

        self.router = Router()
        self.di.define(Router, self.router)

        self.settings = settings
        self.di.define(Settings, self.settings)

        self.handler = self._process_request

        self._register_modules()

        setup_logger = import_item(self.settings.logging.setup)
        setup_logger(self.settings)

    async def __call__(self, scope, receive, send):
        match scope["type"]:
            case "http" | "websocket":
                with logger.contextualize(request_id=uuid4()):
                    await self._handle_request(scope, receive, send)
            case "lifespan":
                await self._handle_lifespan(scope, receive, send)
            case _:
                raise RuntimeError(f"unknown scope '{scope['type']}'")

    def _register_modules(self):
        self.di.scan()

        self.di.scan(
            from_request_impl,
            param_extractor_impl,
            param_converter_impl,
        )

        self.di.scan(self.settings.application)

        for _iface, impl, _name in self.di.iter_all_services():
            if _is_controller(impl):
                self.router.route(impl)

    async def _initialize_extensions(self):
        for extension_name in self.settings.extensions:
            try:
                extension_module = import_item(extension_name)
            except ImportError:
                raise TypeError(f"Extension '{extension_name}' not found")

            try:
                extension_init = getattr(extension_module, "selva_extension")
            except AttributeError:
                raise TypeError(
                    f"Extension '{extension_name}' is missing the 'selva_extension()' function"
                )

            await maybe_async(extension_init, self.di, self.settings)

    async def _initialize_middleware(self):
        middleware = self.settings.middleware
        if len(middleware) == 0:
            return

        middleware = [import_item(name) for name in middleware]

        if errors := [m for m in middleware if not issubclass(m, Middleware)]:
            mid_classes = [f"{m.__module__}.{m.__qualname__}" for m in errors]
            mid_class_name = f"{Middleware.__module__}.{Middleware.__qualname__}"
            raise TypeError(
                f"Middleware classes must be of type '{mid_class_name}': {mid_classes}"
            )

        for cls in reversed(middleware):
            mid = await self.di.get(cls)
            chain = functools.partial(mid, self.handler)
            self.handler = chain

    async def _lifespan_startup(self):
        await self.di._run_startup()
        await self._initialize_extensions()
        await self._initialize_middleware()

    async def _lifespan_shutdown(self):
        await self.di._run_finalizers()

    async def _handle_lifespan(self, _scope, receive, send):
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                logger.trace("Handling lifespan startup")
                try:
                    await self._lifespan_startup()
                    logger.trace("Lifespan startup complete")
                    await send({"type": "lifespan.startup.complete"})
                except Exception as err:
                    logger.trace("Lifespan startup failed")
                    await send({"type": "lifespan.startup.failed", "message": str(err)})
            elif message["type"] == "lifespan.shutdown":
                logger.trace("Handling lifespan shutdown")
                try:
                    await self._lifespan_shutdown()
                    logger.trace("Lifespan shutdown complete")
                    await send({"type": "lifespan.shutdown.complete"})
                except Exception as err:
                    logger.trace("Lifespan shutdown failed")
                    await send(
                        {"type": "lifespan.shutdown.failed", "message": str(err)}
                    )
                break

    async def _handle_request(self, scope, receive, send):
        request = Request(scope, receive, send)

        logger.trace(
            "Started handling of request '{} {} {}'",
            request.method,
            request.path,
            request.raw_query,
        )

        try:
            try:
                await self.handler(request)
            except Exception as err:
                if handler := await self.di.get(
                    ExceptionHandler[type(err)], optional=True
                ):
                    logger.trace(
                        "Handling exception with handler {}.{}",
                        handler.__class__.__module__,
                        handler.__class__.__qualname__,
                    )
                    await handler.handle_exception(request, err)
                else:
                    raise
        except WebSocketException as err:
            await request.websocket.close(err.code, err.reason)
        except (WebSocketDisconnectError, WebSocketError):
            logger.exception("WebSocket error")
            await request.websocket.close()
        except HTTPException as err:
            if websocket := request.websocket:
                logger.error("WebSocket request raise HTTPException")
                await websocket.close()
            else:
                response = request.response
                if response.is_started:
                    logger.error("Response has already started")
                    await response.end()
                    return

                if response.is_finished:
                    logger.error("Response is finished")
                    return

                await respond_text(response, str(err), status=err.status)
        except Exception as err:
            logger.exception("Error processing request")
            await respond_text(
                request.response, str(err), status=HTTPStatus.INTERNAL_SERVER_ERROR
            )

    async def _process_request(self, request: Request):
        method = request.method
        path = request.path

        match = self.router.match(method, path)

        if not match:
            raise HTTPNotFoundException()

        controller = match.route.controller
        action = match.route.action
        path_params = match.params
        request["path_params"] = path_params

        request_params = await self._params_from_request(
            request, match.route.request_params
        )

        instance = await self.di.get(controller)
        await action(instance, request, **request_params)

        response = request.response

        if ws := request.websocket:
            if ws.state != WebSocket.State.CLOSED:
                logger.warning("closing websocket")
                await ws.close()
        elif not response.is_started:
            await respond_status(response, HTTPStatus.INTERNAL_SERVER_ERROR)
        elif not response.is_finished:
            await response.end()

    async def _params_from_request(
        self,
        request: Request,
        params: dict[str, type],
    ) -> dict[str, Any]:
        result = {}

        for name, (param_type, extractor_param) in params.items():
            if extractor_param:
                if inspect.isclass(extractor_param):
                    extractor_type = extractor_param
                else:
                    extractor_type = type(extractor_param)

                extractor = await self.di.get(
                    ParamExtractor[extractor_type], optional=True
                )
                if not extractor:
                    raise MissingRequestParamExtractorImplError(extractor_type)

                param = extractor.extract(request, name, extractor_param)
                if not param:
                    continue

                if converter := await self._find_param_converter(
                    param_type, ParamConverter
                ):
                    value = await maybe_async(converter.from_str, param)
                    result[name] = value
                else:
                    raise MissingParamConverterImplError(param_type)
            else:
                if converter := await self._find_param_converter(
                    param_type, FromRequest
                ):
                    value = await maybe_async(
                        converter.from_request, request, param_type, name
                    )
                    result[name] = value
                else:
                    raise MissingFromRequestImplError(param_type)

        return result

    async def _find_param_converter(
        self, param_type: type, converter_type: type
    ) -> Any | None:
        if typing.get_origin(param_type) is list:
            (param_type,) = typing.get_args(param_type)
            is_generic = True
        else:
            is_generic = False

        for base_type in get_base_types(param_type):
            search_type = list[base_type] if is_generic else base_type
            if converter := await self.di.get(
                converter_type[search_type], optional=True
            ):
                return converter

        return None



from importlib import import_module

__all__ = ("import_item",)


def import_item(name: str):
    """Import a module or an item within a module using its name

    :param name: The name of the item to import, in the format 'package.module.item'.

    :return: The imported module or item within the module
    """

    try:
        return import_module(name)
    except ImportError as err:
        match name.rsplit(".", 1):
            case [module_name, item_name]:
                module = import_module(module_name)
                if item := getattr(module, item_name, None):
                    return item

                raise ImportError(
                    f"module '{module.__name__}' does not have item '{item_name}'"
                )

            case _:
                raise err



import inspect
import typing
from collections.abc import Callable, Iterable
from types import NoneType, UnionType
from typing import Annotated, Any, Optional, TypeVar, Union

from loguru import logger

from selva.di.error import (
    FactoryMissingReturnTypeError,
    InvalidServiceTypeError,
    NonInjectableTypeError,
    TypeVarInGenericServiceError,
)
from selva.di.inject import Inject
from selva.di.service.model import InjectableType, ServiceDependency, ServiceSpec

DI_INITIALIZER = "initialize"
DI_FINALIZER = "finalize"


def _check_optional(type_hint: type, default: Any) -> tuple[type, bool]:
    is_optional = default is None

    if typing.get_origin(type_hint) is UnionType:
        type_args = list(typing.get_args(type_hint))
        if NoneType in type_args:
            type_args.remove(NoneType)

            type_hint = type_args[0]
            is_optional = True

    if typing.get_origin(type_hint) is Union:
        type_arg = typing.get_args(type_hint)[0]
        if type_hint == Optional[type_arg]:
            type_hint = type_arg
            is_optional = True

    return type_hint, is_optional


def _get_injectable_params(hint) -> tuple[type, Any] | None:
    if typing.get_origin(hint) is not Annotated:
        return None

    arg_type, arg_meta = typing.get_args(hint)[0:2]
    if not isinstance(arg_meta, Inject) and arg_meta is not Inject:
        return None

    return arg_type, arg_meta


def _get_service_signature(service: InjectableType) -> Iterable[tuple[str, type, Any]]:
    if inspect.isclass(service):
        for name, hint in typing.get_type_hints(service, include_extras=True).items():
            if params := _get_injectable_params(hint):
                arg_type, arg_meta = params
                yield name, arg_type, arg_meta
    elif inspect.isfunction(service):
        for name, param in inspect.signature(service).parameters.items():
            hint = param.annotation
            if params := _get_injectable_params(hint):
                arg_type, arg_meta = params
                yield name, arg_type, arg_meta
            else:
                yield name, hint, None
    else:
        raise InvalidServiceTypeError(service)


def get_dependencies(
    service: InjectableType,
) -> Iterable[tuple[str, ServiceDependency]]:
    for name, hint, meta in _get_service_signature(service):
        if isinstance(meta, Inject):
            service_name = meta.name
            meta = inspect.Parameter.empty
        else:
            service_name = None

        hint, is_optional = _check_optional(hint, meta)

        dependency = ServiceDependency(hint, name=service_name, optional=is_optional)
        yield name, dependency


def _parse_definition_class(
    service: type, provides: type | None
) -> tuple[type, Callable, Callable]:
    if provides and typing.get_origin(provides):
        if any(isinstance(a, TypeVar) for a in typing.get_args(provides)):
            raise TypeVarInGenericServiceError(provides)

    provided_service = provides or service

    initializer = getattr(service, DI_INITIALIZER, None)
    finalizer = getattr(service, DI_FINALIZER, None)

    return provided_service, initializer, finalizer


def _parse_definition_factory(service: Callable) -> type:
    service_type = typing.get_type_hints(service).get("return")
    if service_type is None:
        raise FactoryMissingReturnTypeError(service)

    provided_service = service_type

    return provided_service


def parse_service_spec(
    injectable: InjectableType,
    provides: type = None,
    name: str = None,
) -> ServiceSpec:
    if inspect.isclass(injectable):
        provided_service, initializer, finalizer = _parse_definition_class(
            injectable, provides
        )

        service = injectable
        factory = None
    elif inspect.isfunction(injectable):
        if provides:
            logger.warning("option 'provides' on a factory function has no effect")

        provided_service = _parse_definition_factory(injectable)
        initializer = None
        finalizer = None

        service = provided_service
        factory = injectable
    else:
        raise NonInjectableTypeError(injectable)

    dependencies = list(get_dependencies(injectable))

    return ServiceSpec(
        service=service,
        provides=provided_service,
        factory=factory,
        name=name,
        dependencies=dependencies,
        initializer=initializer,
        finalizer=finalizer,
    )



default_settings = {
    "application": "application",
    "extensions": [],
    "middleware": [],
    "logging": {
        "setup": "selva.logging.setup.setup_logger",
        "root": "WARNING",
        "level": {},
        "enable": [],
        "disable": [],
    },
    "templates": {
        "backend": None,
        "paths": ["resources/templates"],
        "jinja": {},
    },
    "data": {
        "redis": {},
        "sqlalchemy": {},
    },
}



from collections.abc import Callable
from types import FunctionType
from typing import NamedTuple

InjectableType = type | FunctionType


class ServiceInfo(NamedTuple):
    provides: type | None
    name: str | None
    startup: bool = False


class ServiceDependency(NamedTuple):
    service: type
    name: str | None
    optional: bool = False


class ServiceSpec(NamedTuple):
    service: type
    provides: type
    factory: FunctionType | None
    name: str | None
    dependencies: list[tuple[str, ServiceDependency]]
    initializer: Callable = None
    finalizer: Callable = None



from abc import ABC, abstractmethod
from http import HTTPStatus

from asgikit.responses import Response


class Template(ABC):
    @abstractmethod
    async def respond(
        self,
        response: Response,
        template_name: str,
        context: dict,
        *,
        status: HTTPStatus = HTTPStatus.OK,
        content_type: str = None,
        stream: bool = False,
    ):
        raise NotImplementedError()

    @abstractmethod
    async def render(self, template_name: str, context: dict) -> str:
        raise NotImplementedError()

    @abstractmethod
    async def render_str(self, source: str, context: dict) -> str:
        raise NotImplementedError()



import sys
from functools import cache

from loguru import logger

from selva.configuration.settings import Settings
from selva.logging.stdlib import setup_loguru_std_logging_interceptor


def setup_logger(settings: Settings):
    enable = [(name, True) for name in settings.logging.enable]
    disable = [(name, False) for name in settings.logging.disable]

    # enabling has precedence over disabling
    # therefore the "enable" list comes after the "disable" list
    activation = disable + enable

    log_config = settings.get("logging", {})
    root_level = logger.level(log_config.get("root", "WARNING"))
    log_level = {
        name: logger.level(value) for name, value in log_config.get("level", {}).items()
    }

    filter_func = filter_func_factory(root_level, log_level)
    handler = {"sink": sys.stderr, "filter": filter_func}

    logger.configure(handlers=[handler], activation=activation)

    setup_loguru_std_logging_interceptor()


def filter_func_factory(root_level, log_level: dict):
    @cache
    def has_level(name: str, record_level):
        level = log_level.get(name)

        while not level:
            match name.rsplit(".", 1):
                case [first, _last]:
                    name = first
                    level = log_level.get(name)
                case _:
                    level = root_level

        return record_level.no >= level.no

    def filter_func(record):
        name = record["name"]
        record_level = record["level"]
        return has_level(name, record_level)

    return filter_func



import typing
from typing import Any, Generic, TypeVar

from pydantic import GetCoreSchemaHandler, ValidatorFunctionWrapHandler
from pydantic_core import PydanticCustomError, core_schema

from selva._util.import_item import import_item

T = TypeVar("T")


class DottedPath(Generic[T]):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        origin = typing.get_origin(source_type)
        if origin is None:  # used as `x: Owner` without params
            item_tp = Any
        else:
            item_tp = typing.get_args(source_type)[0]

        item_schema = handler.generate_schema(Any)

        def validate_from_str(
            value: str, function_handler: ValidatorFunctionWrapHandler
        ) -> item_tp:
            try:
                item = import_item(value)
            except ImportError as e:
                raise PydanticCustomError(
                    "invalid_dotted_path",
                    "unable to import '{path}': {error}",
                    {"path": value, "error": e.msg},
                ) from e

            return function_handler(item)

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_wrap_validator_function(
                    validate_from_str, item_schema
                ),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_subclass_schema(str),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: f"{instance.__module__}.{instance.__qualname__}",
                when_used="json-unless-none",
            ),
        )



from typing import Protocol, Type, TypeVar, runtime_checkable

from asgikit.requests import Request

T = TypeVar("T")


@runtime_checkable
class ParamExtractor(Protocol[T]):
    def extract(self, request: Request, parameter_name: str, metadata: T | Type[T]):
        raise NotImplementedError()



import inspect
from typing import Type

from asgikit.requests import Request

from selva.web.converter.decorator import register_param_extractor
from selva.web.converter.error import PathParamNotFoundError

__all__ = (
    "FromPath",
    "FromQuery",
    "FromHeader",
    "FromCookie",
)


class FromRequestParam:
    name: str = None

    def __init__(self, name: str = None):
        self.name = name


class FromPath(FromRequestParam):
    pass


@register_param_extractor(FromPath)
class FromPathExtractor:
    @staticmethod
    def extract(
        request: Request,
        parameter_name: str,
        metadata: FromPath | Type[FromPath],
    ) -> str:
        if inspect.isclass(metadata):
            name = parameter_name
        else:
            name = metadata.name or parameter_name

        param = request["path_params"].get(name)
        if not param:
            raise PathParamNotFoundError(name)

        return param


class FromQuery(FromRequestParam):
    pass


@register_param_extractor(FromQuery)
class FromQueryExtractor:
    @staticmethod
    def extract(
        request: Request,
        parameter_name: str,
        metadata: FromQuery | Type[FromQuery],
    ) -> str:
        if inspect.isclass(metadata):
            name = parameter_name
        else:
            name = metadata.name or parameter_name

        return request.query.get(name)


class FromHeader(FromRequestParam):
    pass


@register_param_extractor(FromHeader)
class FromHeaderExtractor:
    @staticmethod
    def extract(
        request: Request,
        parameter_name: str,
        metadata: FromHeader | Type[FromHeader],
    ) -> str | None:
        if inspect.isclass(metadata):
            name = parameter_name
        else:
            name = metadata.name or parameter_name

        candidate_names = (
            name,
            name.lower(),
            "-".join(name.split("_")),
            "-".join(i.capitalize() for i in name.split("_")),
        )

        for cand in candidate_names:
            if value := request.headers.get(cand):
                return value

        return None


class FromCookie(FromRequestParam):
    pass


@register_param_extractor(FromCookie)
class FromCookieExtractor:
    @staticmethod
    def extract(
        request: Request,
        parameter_name: str,
        metadata: FromCookie | Type[FromCookie],
    ) -> str:
        if inspect.isclass(metadata):
            name = parameter_name
        else:
            name = metadata.name or parameter_name

        return request.cookie.get(name)



from typing import Protocol, Type, TypeVar, runtime_checkable

from asgikit.requests import Request

from selva.di.decorator import service

TExc = TypeVar("TExc", bound=BaseException)


@runtime_checkable
class ExceptionHandler(Protocol[TExc]):
    async def handle_exception(self, request: Request, exc: TExc):
        raise NotImplementedError()


def exception_handler(exc: Type[BaseException]):
    def inner(cls):
        assert issubclass(cls, ExceptionHandler)
        return service(cls, provides=ExceptionHandler[exc])

    return inner



import inspect
from collections.abc import Callable
from enum import Enum
from http import HTTPMethod
from typing import NamedTuple

from asgikit.requests import Request

from selva.di import service

CONTROLLER_ATTRIBUTE = "__selva_web_controller__"
ACTION_ATTRIBUTE = "__selva_web_action__"


class ControllerInfo(NamedTuple):
    path: str


class ActionType(Enum):
    GET = HTTPMethod.GET
    HEAD = HTTPMethod.HEAD
    POST = HTTPMethod.POST
    PUT = HTTPMethod.PUT
    PATCH = HTTPMethod.PATCH
    DELETE = HTTPMethod.DELETE
    OPTIONS = HTTPMethod.OPTIONS
    WEBSOCKET = None

    @property
    def is_websocket(self) -> bool:
        return self is ActionType.WEBSOCKET


class ActionInfo(NamedTuple):
    type: ActionType
    path: str


def controller(target: type | str):
    if inspect.isclass(target):
        path = ""
        cls = target
    elif isinstance(target, str):
        path = target.strip("/")
        cls = None
    else:
        raise TypeError(f"@controller must be applied to class, '{target}' given")

    def inner(arg: type):
        setattr(arg, CONTROLLER_ATTRIBUTE, ControllerInfo(path))
        return service(arg)

    return inner(cls) if cls else inner


def route(action: Callable = None, /, *, method: HTTPMethod | None, path: str | None):
    path = path.strip("/") if path else ""

    def inner(arg: Callable):
        if not inspect.iscoroutinefunction(arg):
            raise TypeError("Handler method must be async")

        params = list(inspect.signature(arg).parameters.values())
        if len(params) < 2:
            raise TypeError(
                "Handler method must have at least 'self' and 'request' parameters"
            )

        req_param = params[1].annotation
        if req_param is not inspect.Signature.empty and req_param is not Request:
            raise TypeError(
                f"Handler request parameter must be of type "
                f"'{Request.__module__}.{Request.__name__}'"
            )

        if any(p.annotation is inspect.Signature.empty for p in params[2:]):
            raise TypeError("Handler parameters must be typed")

        setattr(arg, ACTION_ATTRIBUTE, ActionInfo(ActionType(method), path))
        return arg

    return inner(action) if action else inner


def _route(method: HTTPMethod | None, path_or_action: str | Callable):
    if isinstance(path_or_action, str):
        path = path_or_action.strip("/")
        action = None
    else:
        path = ""
        action = path_or_action

    return route(action, method=method, path=path)


def get(path_or_action: str | Callable):
    return _route(HTTPMethod.GET, path_or_action)


def post(path_or_action: str | Callable):
    return _route(HTTPMethod.POST, path_or_action)


def put(path_or_action: str | Callable):
    return _route(HTTPMethod.PUT, path_or_action)


def patch(path_or_action: str | Callable):
    return _route(HTTPMethod.PATCH, path_or_action)


def delete(path_or_action: str | Callable):
    return _route(HTTPMethod.DELETE, path_or_action)


def websocket(path_or_action: str | Callable):
    return _route(None, path_or_action)



from collections.abc import Awaitable, Callable
from typing import Protocol, runtime_checkable

from asgikit.requests import Request

__all__ = ("Middleware", "CallNext")


CallNext = Callable[[Request], Awaitable]


@runtime_checkable
class Middleware(Protocol):
    async def __call__(
        self,
        call_next: CallNext,
        request: Request,
    ):
        raise NotImplementedError()



from typing import Type

from selva.di.decorator import service
from selva.web.converter.from_request import FromRequest
from selva.web.converter.param_converter import ParamConverter
from selva.web.converter.param_extractor import ParamExtractor


def register_from_request(target: Type):
    def inner(cls):
        assert issubclass(cls, FromRequest)
        return service(cls, provides=FromRequest[target])

    return inner


def register_param_converter(target: Type):
    def inner(cls):
        if not hasattr(cls, ParamConverter.into_str.__name__):
            setattr(cls, ParamConverter.into_str.__name__, ParamConverter.into_str)

        assert issubclass(cls, ParamConverter)
        return service(cls, provides=ParamConverter[target])

    return inner


def register_param_extractor(target: Type):
    def inner(cls):
        assert issubclass(cls, ParamExtractor)
        return service(cls, provides=ParamExtractor[target])

    return inner



from selva.web.converter.from_request import FromRequest
from selva.web.converter.param_converter import ParamConverter
from selva.web.converter.param_extractor import ParamExtractor


class MissingFromRequestImplError(Exception):
    def __init__(self, param_type):
        super().__init__(
            f"no implementation of '{FromRequest.__name__}' found for type {param_type}"
        )
        self.param_type = param_type


class MissingParamConverterImplError(Exception):
    def __init__(self, param_type):
        super().__init__(
            f"no implementation of '{ParamConverter.__name__}' found for type {param_type}"
        )
        self.param_type = param_type


class MissingRequestParamExtractorImplError(Exception):
    def __init__(self, param_type):
        super().__init__(
            f"no implementation of '{ParamExtractor.__name__}' found for type {param_type}"
        )
        self.param_type = param_type


class PathParamNotFoundError(Exception):
    def __init__(self, name: str):
        super().__init__(f"path parameter '{name}' not found")
        self.name = name



from collections.abc import Awaitable
from typing import Protocol, TypeVar, runtime_checkable

__all__ = ("ParamConverter",)

T = TypeVar("T")


@runtime_checkable
class ParamConverter(Protocol[T]):
    """Convert values from and to request parameters

    Request parameters come from path, querystring or headers.
    Request parameters can be extended with `ParamExtractor` implementations
    """

    def from_str(self, data: str) -> T | Awaitable[T]:
        raise NotImplementedError()

    def into_str(self, value: T) -> str | Awaitable[str]:
        return str(value)



from collections.abc import Callable
from types import ModuleType
from typing import Any, Literal, Self, Type

from pydantic import BaseModel, ConfigDict, Field, model_validator
from sqlalchemy import URL, make_url
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.session import JoinTransactionMode

from selva._util.pydantic import DottedPath


class SqlAlchemyExecutionOptions(BaseModel):
    """SQLAlchemy execution options.

    Defined in https://docs.sqlalchemy.org/en/20/core/connections.html#sqlalchemy.engine.Connection.execution_options
    """

    model_config = ConfigDict(extra="forbid")

    logging_token: str = None
    isolation_level: str = None
    no_parameters: bool = None
    stream_results: bool = None
    max_row_buffer: int = None
    yield_per: int = None
    insertmanyvalues_page_size: int = None
    schema_translate_map: dict[str | None, str] = None


class SqlAlchemyOptions(BaseModel):
    """SQLAlchemy options

    Defined in https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.create_engine
    """

    model_config = ConfigDict(extra="forbid")

    connect_args: dict[str, Any] = None
    echo: bool = None
    echo_pool: bool = None
    enable_from_linting: bool = None
    execution_options: SqlAlchemyExecutionOptions = None
    hide_parameters: bool = None
    insertmanyvalues_page_size: int = None
    isolation_level: str = None
    json_deserializer: DottedPath[Callable] = None
    json_serializer: DottedPath[Callable] = None
    label_length: int = None
    logging_name: str = None
    max_identifier_length: int = None
    max_overflow: int = None
    module: DottedPath[ModuleType] = None
    paramstyle: Literal["qmark", "numeric", "named", "format", "pyformat"] = None
    poolclass: DottedPath = None
    pool_logging_name: str = None
    pool_pre_ping: bool = None
    pool_size: int = None
    pool_recycle: int = None
    pool_reset_on_return: Literal["rollback", "commit"] = None
    pool_timeout: int = None
    pool_use_lifo: bool = None
    plugins: list[str] = None
    query_cache_size: int = None
    use_insertmanyvalues: bool = None


class SqlAlchemyEngineSettings(BaseModel):
    """Settings for a SQLAlchemy connection defined in a settings file."""

    model_config = ConfigDict(extra="forbid")

    url: str = None
    username: str = None
    password: str = None
    drivername: str = None
    host: str = None
    port: int = None
    database: str = None
    query: dict[str, str] = None
    options: SqlAlchemyOptions = None

    @model_validator(mode="after")
    def validate(self) -> Self:
        if self.url and (self.drivername or self.host or self.port or self.database):
            raise ValueError(
                "Either 'url' should be provided, or 'drivername', 'host', 'port' and 'database'"
            )

        return self

    def get_url(self) -> URL:
        if url := self.url:
            url = make_url(url)
            if username := self.username:
                url = url.set(username=username)
            if password := self.password:
                url = url.set(password=password)
            if query := self.query:
                url = url.set(query=query)
        else:
            url = URL.create(
                drivername=self.drivername,
                host=self.host,
                port=self.port,
                database=self.database,
                username=self.username,
                password=self.password,
                query=self.query,
            )

        return url


class SqlAlchemySessionOptions(BaseModel):
    class_: DottedPath[Type[AsyncSession]] = Field(default=None, alias="class")
    autoflush: bool = (None,)
    expire_on_commit: bool = (None,)
    autobegin: bool = (None,)
    twophase: bool = (None,)
    enable_baked_queries: bool = (None,)
    info: dict[str, Any] = (None,)
    query_cls: DottedPath[Type[Query[Any]]] = (None,)
    join_transaction_mode: JoinTransactionMode = (None,)
    close_resets_only: bool = (None,)


class SqlAlchemySessionSettings(BaseModel):
    """Settings for the SQLAlchemy session defined in a settings file."""

    model_config = ConfigDict(extra="forbid")

    options: SqlAlchemySessionOptions = None
    binds: dict[DottedPath, str] = None


class SqlAlchemySettings(BaseModel):
    """Settings for the SQLAlchemy session defined in a settings file."""

    model_config = ConfigDict(extra="forbid")

    connections: dict[str, SqlAlchemyEngineSettings]
    session: SqlAlchemySessionSettings = Field(
        default_factory=SqlAlchemySessionSettings
    )



from collections.abc import Awaitable
from typing import Any, Protocol, Type, TypeVar, runtime_checkable

from asgikit.requests import Request

__all__ = ("FromRequest",)

T = TypeVar("T")


@runtime_checkable
class FromRequest(Protocol[T]):
    """Base class for services that extract values from the request"""

    def from_request(
        self,
        request: Request,
        original_type: Type[T],
        parameter_name: str,
        metadata: Any = None,
    ) -> T | Awaitable[T]:
        """Extract values from the request based on the type hint

        The `FromRequest` implementation may be for a superclass of the declared
        type, therefore the original type is passed as a parameter

        :param request: The request to extract values from
        :param original_type: Declared type on handler method signature
        :param parameter_name: Name of the parameter on the handler method
        :param metadata: Any metadata associated with the type
        """

        raise NotImplementedError()



import typing
import warnings
from collections import defaultdict

from selva.di.error import ServiceAlreadyRegisteredError, ServiceNotFoundError
from selva.di.service.model import ServiceSpec

__all__ = ("ServiceRecord", "ServiceRegistry")


class ServiceRecord:
    def __init__(self):
        self.providers: dict[str | None, ServiceSpec] = {}

    def add(self, service: ServiceSpec, name: str = None):
        if name in self.providers:
            raise ServiceAlreadyRegisteredError(service.provides, name)

        self.providers[name] = service

    def get(self, name: str = None) -> ServiceSpec | None:
        if service := self.providers.get(name):
            return service

        if default := self.providers.get(None):
            message = (
                f"using default service instead of '{name}'"
                f" for '{default.provides.__qualname__}'"
            )

            warnings.warn(message)
            return default

        return None

    def __contains__(self, name: str | None) -> bool:
        return name in self.providers


def _get_key_with_name(key: type | tuple[type, str]) -> tuple[type, str | None]:
    # typing.get_origin(k) to test Generic[T]
    if isinstance(key, type) or typing.get_origin(key) is not None:
        return key, None

    if isinstance(key, tuple) and len(key) == 2:
        return key

    raise ValueError(key)


class ServiceRegistry:
    def __init__(self):
        self.services: dict[type, ServiceRecord] = defaultdict(ServiceRecord)

    def get(self, key: type, name: str = None) -> ServiceSpec | None:
        try:
            return self[key, name]
        except ServiceNotFoundError:
            return None

    def __getitem__(self, key: type | tuple[type, str]):
        inner_key, name = _get_key_with_name(key)

        if service := self.services[inner_key].get(name):
            return service

        raise ServiceNotFoundError(inner_key, name=name)

    def __setitem__(self, key: type | tuple[type, str], value: ServiceSpec):
        inner_key, name = _get_key_with_name(key)
        self.services[inner_key].add(value, name)

    def __contains__(self, key: type | tuple[type, str]):
        inner_key, name = _get_key_with_name(key)
        return inner_key in self.services and name in self.services[inner_key]



# flake8: noqa: F401
# ruff: noqa: F401

from selva.web.application import Selva
from selva.web.converter.param_extractor_impl import (
    FromCookie,
    FromHeader,
    FromPath,
    FromQuery,
)
from selva.web.middleware import Middleware
from selva.web.routing.decorator import (
    controller,
    delete,
    get,
    patch,
    post,
    put,
    websocket,
)



import re
from collections.abc import Mapping, Sequence
from typing import Any

RE_VARIABLE = re.compile(
    r"""
        \$\{                        # Expression start with '${'
            \ ?                     # Blank space before variable name
            (?P<name>\w+?)          # Environment variable name
            \ ?                     # Blank space after variable name
            (?:                     # Default value non-capture group
                :                   # Collon separating variable name and default value
                \ ?                 # Blank space before default value
                (?P<default>.*?)    # Default value
                \ ?                 # Blank space after default value
            )?                      # End of default value non-capture group
        }                           # Expression end with '}'
    """,
    re.VERBOSE,
)


SELVA_PREFIX = "SELVA__"


def parse_settings_from_env(source: Mapping[str, str]) -> dict:
    result = {}

    for name, value in source.items():
        if not name.startswith(SELVA_PREFIX):
            continue

        current_ref = result
        keys = name.removeprefix(SELVA_PREFIX).split("__")

        match keys:
            case [key] if key.strip() != "":
                current_key = key.lower()
            case [first_key, *keys, last_key]:
                for key in [first_key] + keys:
                    key = key.lower()
                    if key not in current_ref:
                        current_ref[key] = {}
                    current_ref = current_ref[key]
                current_key = last_key.lower()
            case _:
                continue

        current_ref[current_key] = value

    return result


def replace_variables_with_env(settings: str, environ: Mapping[str, str]):
    for match in RE_VARIABLE.finditer(settings):
        name = match.group("name")

        if value := environ.get(name):
            settings = settings.replace(match.group(), value)
            continue

        if value := match.group("default"):
            settings = settings.replace(match.group(), value)
            continue

        raise ValueError(
            f"{name} environment variable is not defined and does not contain a default value"
        )

    return settings


def replace_variables_recursive(
    data: Mapping | Sequence | str | Any, environ: Mapping[str, str]
):
    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = replace_variables_recursive(value, environ)
        return data

    if isinstance(data, list):
        return [replace_variables_recursive(value, environ) for value in data]

    if isinstance(data, str):
        return replace_variables_with_env(data, environ)

    return data



import importlib
import inspect
import pkgutil
from collections.abc import Callable, Iterable
from types import ModuleType
from typing import Any


def _is_class_or_function(arg) -> bool:
    return inspect.isclass(arg) or inspect.isfunction(arg)


def _scan_members(module, predicate):
    for _name, member in inspect.getmembers(module, predicate):
        if member.__module__ == module.__name__:
            yield member


def scan_packages(
    modules_to_scan: Iterable[str | ModuleType],
    predicate: Callable[[Any], bool] = None,
) -> Iterable[type]:
    if predicate and not inspect.isfunction(predicate):
        raise ValueError("invalid predicate")

    for module in modules_to_scan:
        if isinstance(module, str):
            module = importlib.import_module(module)

        def scan_predicate(arg):
            predicate_result = predicate(arg) if predicate else True
            return _is_class_or_function(arg) and predicate_result

        yield from _scan_members(module, scan_predicate)

        spec = getattr(module, "__spec__", None)
        if not spec or not spec.submodule_search_locations:
            # module is not a package
            continue

        search_paths = spec.submodule_search_locations

        prefix = spec.name
        if prefix:
            prefix += "."

        for _module_finder, name, _ispkg in pkgutil.walk_packages(search_paths, prefix):
            submodule = importlib.import_module(name)
            yield from _scan_members(submodule, scan_predicate)



from collections.abc import Callable
from typing import Literal, Type

from jinja2 import BaseLoader, BytecodeCache, Undefined
from pydantic import BaseModel, ConfigDict

from selva._util.pydantic import DottedPath


class JinjaTemplateSettings(BaseModel):
    model_config = ConfigDict(extra="forbid")

    block_start_string: str = None
    block_end_string: str = None
    variable_start_string: str = None
    variable_end_string: str = None
    comment_start_string: str = None
    comment_end_string: str = None
    line_statement_prefix: str = None
    line_comment_prefix: str = None
    trim_blocks: bool = None
    lstrip_blocks: bool = None
    newline_sequence: Literal["\n", "\r\n", "\r"] = None
    keep_trailing_newline: bool = None
    extensions: list[str] = None
    optimized: bool = None
    undefined: DottedPath[Type[Undefined]] = None
    finalize: DottedPath[Callable[..., None]] = None
    autoescape: bool | DottedPath[Callable[[str], bool]] = None
    loader: DottedPath[BaseLoader] = None
    cache_size: int = None
    auto_reload: bool = None
    bytecode_cache: DottedPath[BytecodeCache] = None



from types import NoneType
from typing import Literal, Self, Type

from pydantic import BaseModel, ConfigDict, model_serializer, model_validator
from redis import RedisError

from selva._util.pydantic import DottedPath


class ConstantBackoffSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    backoff: int


class ExponentialBackoffSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    cap: int = None
    base: int = None


class BackoffSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    no_backoff: NoneType = None
    constant: ConstantBackoffSchema = None
    exponential: ExponentialBackoffSchema = None
    full_jitter: ExponentialBackoffSchema = None
    equal_jitter: ExponentialBackoffSchema = None
    decorrelated_jitter: ExponentialBackoffSchema = None

    @model_validator(mode="before")
    def validator(cls, data):
        if len(data) != 1:
            raise ValueError("Only one backoff value can be set")

        return data


class RetrySchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    backoff: BackoffSchema
    retries: int
    supported_errors: tuple[DottedPath[Type[RedisError]], ...] = None


class RedisOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")

    socket_timeout: float = None
    socket_connect_timeout: float = None
    socket_keepalive: bool = None
    socket_keepalive_options: dict[int, int | bytes] = None
    unix_socket_path: str = None
    encoding: str = None
    encoding_errors: Literal["strict", "ignore", "replace"] = None
    decode_responses: bool = None
    retry_on_timeout: bool = None
    retry_on_error: list = None
    ssl: bool = None
    ssl_keyfile: str = None
    ssl_certfile: str = None
    ssl_cert_reqs: str = None
    ssl_ca_certs: str = None
    ssl_ca_data: str = None
    ssl_check_hostname: bool = None
    max_connections: int = None
    single_connection_client: bool = None
    health_check_interval: int = None
    client_name: str = None
    lib_name: str = None
    lib_version: str = None
    retry: RetrySchema = None
    auto_close_connection_pool: bool = None
    protocol: int = None


class RedisSettings(BaseModel):
    model_config = ConfigDict(extra="forbid")

    url: str = None
    host: str = None
    port: int = None
    db: int = None
    username: str = None
    password: str = None
    options: RedisOptions = None

    @model_validator(mode="after")
    def validator(self) -> Self:
        if self.url and (self.host or self.port or (self.db is not None)):
            raise ValueError(
                "Either 'url' should be provided, or 'host', 'port' and 'db'"
            )

        return self

    @model_serializer(when_used="unless-none")
    def serializer(self):
        data = {
            field: getattr(self, field)
            for field in set(self.model_fields_set)
            if field != "options"
        }

        if self.options:
            data |= self.options.model_dump(exclude_unset=True)

        return data



from importlib.util import find_spec

from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.di.decorator import service as service_decorator

from .service import make_engine_service  # noqa: F401
from .service import engine_dict_service, sessionmaker_service


def selva_extension(container: Container, settings: Settings):
    if find_spec("sqlalchemy") is None:
        return

    for name in settings.data.sqlalchemy.connections:
        service_name = name if name != "default" else None

        container.register(
            service_decorator(make_engine_service(name), name=service_name)
        )

    container.register(service_decorator(engine_dict_service))
    container.register(service_decorator(sessionmaker_service))



def logfmt_format(record):
    message = {
        "time": f"{record['time']}",
        "level": f"{record['level']}",
        "source": f"{record['module']}.{record['name']}:{record['function']}:{record['line']}",
        "message": '"' + record["message"].replace('"', r"\"") + '"',
    }
    message |= record["extra"]

    return " ".join(f"{key}={value}" for key, value in message.items()) + "\n"



import inspect
import logging

from loguru import logger


class InterceptHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def setup_loguru_std_logging_interceptor():
    logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)



from importlib.util import find_spec

from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.di.decorator import service as service_decorator

from .service import make_service


def selva_extension(container: Container, settings: Settings):
    if find_spec("redis") is None:
        return

    for name in settings.data.redis:
        service_name = name if name != "default" else None

        container.register(service_decorator(make_service(name), name=service_name))



from loguru import logger
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine

from selva.configuration.settings import Settings
from selva.di import Container, Inject
from selva.ext.data.sqlalchemy.settings import SqlAlchemyEngineSettings

from .settings import SqlAlchemySettings


def make_engine_service(name: str):
    async def engine_service(
        settings: Settings,
    ) -> AsyncEngine:
        sa_settings = SqlAlchemyEngineSettings.model_validate(
            settings.data.sqlalchemy.connections[name]
        )
        url = sa_settings.get_url()

        if sa_options := sa_settings.options:
            options = sa_options.model_dump(exclude_unset=True)
        else:
            options = {}

        engine = create_async_engine(url, **options)
        yield engine
        await engine.dispose()

    return engine_service


async def engine_dict_service(
    di: Container, settings: Settings
) -> dict[str, AsyncEngine]:
    return {
        db: await di.get(AsyncEngine, name=db if db != "default" else None)
        for db in settings.data.sqlalchemy.connections
    }


async def sessionmaker_service(
    engines: dict[str, AsyncEngine],
    settings: Settings,
) -> async_sessionmaker:
    sqlalchemy_settings = SqlAlchemySettings.model_validate(settings.data.sqlalchemy)
    if session_options := sqlalchemy_settings.session.options:
        options = session_options.model_dump(exclude_unset=True)
    else:
        options = {}

    if binds := sqlalchemy_settings.session.binds:
        binds_config = {}
        for mapper, engine_name in binds.items():
            if engine := engines.get(engine_name):
                binds_config[mapper] = engine
            else:
                raise ValueError(f"No engine with name '{engine_name}'")

        return async_sessionmaker(binds=binds_config, **options)
    else:
        engine = engines.get("default")
        if not engine:
            name, engine = next(iter(engines.items()))
            logger.warning("Using connection '{}' for sqlalchemy session", name)

        return async_sessionmaker(engine, **options)



from typing import Protocol, runtime_checkable


@runtime_checkable
class Interceptor(Protocol):
    async def intercept(self, instance: object, service_type: type):
        raise NotImplementedError()



from functools import cache
from typing import Generic, Protocol, get_origin

__all__ = ("get_base_types",)


@cache
def get_base_types(base_class: type) -> list[type | Generic]:
    if origin := get_origin(base_class):
        result = [base_class] + list(origin.mro())
        base_class = origin
    else:
        result = list(base_class.mro())

    # remove 'object' type
    result.pop()

    if Generic in result:
        if Protocol in result:
            index = result.index(Protocol)
        else:
            index = result.index(Generic)

        result = result[:index]

    for origin, generic in _get_generic_origins(base_class):
        if origin in result:
            index = result.index(origin)
            result.insert(index, generic)

    return result


def _get_generic_origins(base_class):
    if orig_bases := getattr(base_class, "__orig_bases__", None):
        for generic in orig_bases:
            origin = get_origin(generic)
            yield origin, generic
            yield from _get_generic_origins(origin)



from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.web.templates import Template


async def selva_extension(container: Container, settings: Settings):
    backend = settings.templates.backend

    if backend and backend != "jinja":
        return

    if not backend and (current := await container.get(Template, optional=True)):
        cls = current.__class__
        current_class = f"{cls.__module__}.{cls.__qualname__}"

        raise ValueError(
            f"Template backend already registered with '{current_class}'. "
            "Please define `templates.backend` property."
        )

    container.scan(f"{__package__}.service")



import copy
import os
from collections.abc import Mapping
from copy import deepcopy
from functools import cache
from pathlib import Path
from typing import Any

from loguru import logger
from ruamel.yaml import YAML

from selva.configuration.defaults import default_settings
from selva.configuration.environment import (
    parse_settings_from_env,
    replace_variables_recursive,
)

__all__ = ("Settings", "SettingsError", "get_settings")

SETTINGS_DIR_ENV = "SELVA_SETTINGS_DIR"
SETTINGS_FILE_ENV = "SELVA_SETTINGS_FILE"

DEFAULT_SETTINGS_DIR = str(Path("configuration"))
DEFAULT_SETTINGS_FILE = "settings.yaml"

SELVA_PROFILE = "SELVA_PROFILE"


class Settings(Mapping[str, Any]):
    def __init__(self, data: dict):
        for key, value in data.items():
            if isinstance(value, dict):
                data[key] = Settings(value)

        self.__data = data

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self):
        return iter(self.__data)

    def __contains__(self, key: str):
        return key in self.__data

    def __getitem__(self, key: str):
        return self.__data[key]

    def __getattr__(self, item: str):
        try:
            return self.__data[item]
        except KeyError:
            raise AttributeError(item)

    def __copy__(self):
        return Settings(copy.copy(self.__data))

    def __deepcopy__(self, memodict):
        data = copy.deepcopy(self.__data, memodict)
        return Settings(data)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Settings):
            return self.__data == other.__data
        if isinstance(other, Mapping):
            return self.__data == other

        return False

    def __str__(self):
        return str(self.__data)

    def __repr__(self):
        return repr(self.__data)


class SettingsError(Exception):
    def __init__(self, path: Path):
        super().__init__(f"cannot load settings from {path}")
        self.path = path


@cache
def get_settings() -> Settings:
    return _get_settings_nocache()


def _get_settings_nocache() -> Settings:
    # get default settings
    settings = deepcopy(default_settings)

    # merge with main settings file (settings.yaml)
    merge_recursive(settings, get_settings_for_profile())

    # merge with environment settings file (settings_$SELVA_PROFILE.yaml)
    if active_profile := os.getenv(SELVA_PROFILE):
        merge_recursive(settings, get_settings_for_profile(active_profile))

    # merge with environment variables (SELVA_*)
    from_env_vars = parse_settings_from_env(os.environ)
    merge_recursive(settings, from_env_vars)

    settings = replace_variables_recursive(settings, os.environ)
    return Settings(settings)


def get_settings_for_profile(profile: str = None) -> dict[str, Any]:
    settings_file = os.getenv(SETTINGS_FILE_ENV, DEFAULT_SETTINGS_FILE)
    settings_dir_path = Path(os.getenv(SETTINGS_DIR_ENV, DEFAULT_SETTINGS_DIR))
    settings_file_path = settings_dir_path / settings_file

    if profile:
        settings_file_path = settings_file_path.with_stem(
            f"{settings_file_path.stem}_{profile}"
        )

    settings_file_path = settings_file_path.absolute()

    try:
        logger.info("settings loaded from {}", settings_file_path)
        yaml = YAML(typ="safe")
        return yaml.load(settings_file_path) or {}
    except FileNotFoundError:
        if profile:
            logger.warning(
                "no settings file found for profile '{}' at {}",
                profile,
                settings_file_path,
            )

        return {}
    except Exception as err:
        raise SettingsError(settings_file_path) from err


def merge_recursive(destination: dict, source: dict):
    for key in source:
        if key in destination and all(
            isinstance(arg[key], dict) for arg in (destination, source)
        ):
            merge_recursive(destination[key], source[key])
        else:
            destination[key] = deepcopy(source[key])



import asyncio
import inspect
from collections.abc import AsyncGenerator, Awaitable, Generator, Iterable
from types import FunctionType, ModuleType
from typing import Any, Type, TypeVar

from loguru import logger

from selva._util.maybe_async import maybe_async
from selva._util.package_scan import scan_packages
from selva.di.decorator import DI_ATTRIBUTE_SERVICE
from selva.di.decorator import service as service_decorator
from selva.di.error import (
    DependencyLoopError,
    NonInjectableTypeError,
    ServiceNotFoundError,
    ServiceWithoutDecoratorError,
)
from selva.di.interceptor import Interceptor
from selva.di.service.model import InjectableType, ServiceDependency, ServiceSpec
from selva.di.service.parse import get_dependencies, parse_service_spec
from selva.di.service.registry import ServiceRegistry

T = TypeVar("T")


class Container:
    def __init__(self):
        self.registry = ServiceRegistry()
        self.store: dict[tuple[type, str | None], Any] = {}
        self.finalizers: list[Awaitable] = []
        self.startup: list[tuple[Type, str | None]] = []
        self.interceptors: list[Type[Interceptor]] = []

    def register(self, injectable: InjectableType):
        service_info = getattr(injectable, DI_ATTRIBUTE_SERVICE, None)

        if not service_info:
            if inspect.isfunction(injectable) or inspect.isclass(injectable):
                raise ServiceWithoutDecoratorError(injectable)

            raise NonInjectableTypeError(injectable)

        provides, name, startup = service_info
        service_spec = parse_service_spec(injectable, provides, name)
        provided_service = service_spec.provides

        self.registry[provided_service, name] = service_spec

        if startup:
            self.startup.append((service_spec.provides, name))

        if provides:
            logger.trace(
                "service registered: {}.{}; provides={}.{} name={}",
                injectable.__module__,
                injectable.__qualname__,
                provides.__module__,
                provides.__qualname__,
                name or "",
            )
        else:
            logger.trace(
                "service registered: {}.{}; name={}",
                injectable.__module__,
                injectable.__qualname__,
                name or "",
            )

    def define(self, service_type: type, instance: Any, *, name: str = None):
        self.store[service_type, name] = instance

        logger.trace(
            "service defined: {}; name={}", service_type.__qualname__, name or ""
        )

    def interceptor(self, interceptor: Type[Interceptor]):
        self.register(
            service_decorator(
                interceptor,
                provides=Interceptor,
                name=f"{interceptor.__module__}.{interceptor.__qualname__}",
            )
        )
        self.interceptors.append(interceptor)

        logger.trace("interceptor registered: {}", interceptor.__qualname__)

    def scan(self, *packages: str | ModuleType):
        def predicate_services(item: Any):
            return hasattr(item, DI_ATTRIBUTE_SERVICE)

        for found_service in scan_packages(packages, predicate_services):
            self.register(found_service)

    def has(self, service_type: type, name: str = None) -> bool:
        definition = self.registry.get(service_type, name=name)
        return definition is not None

    def iter_service(
        self, key: type
    ) -> Iterable[tuple[type | FunctionType, str | None]]:
        record = self.registry.services.get(key)
        if not record:
            raise ServiceNotFoundError(key)

        for name, definition in record.providers.items():
            yield definition.service, name

    def iter_all_services(
        self,
    ) -> Iterable[tuple[type, type | FunctionType | None, str | None]]:
        for interface, record in self.registry.services.items():
            for name, definition in record.providers.items():
                yield interface, definition.service, name

    async def get(
        self, service_type: Type[T], *, name: str = None, optional=False
    ) -> T:
        dependency = ServiceDependency(service_type, name=name, optional=optional)
        return await self._get(dependency)

    def _get_from_cache(self, service_type: Type[T], name: str | None) -> T | None:
        if instance := self.store.get((service_type, name)):
            return instance

        return None

    async def _get(
        self,
        dependency: ServiceDependency,
        stack: list = None,
    ) -> Any | None:
        service_type, name = dependency.service, dependency.name

        # check if service exists in cache
        if instance := self._get_from_cache(service_type, name):
            return instance

        try:
            service_spec = self.registry.get(service_type, name)
            if not service_spec:
                raise ServiceNotFoundError(service_type, name)
        except ServiceNotFoundError:
            if dependency.optional:
                return None
            raise

        stack = stack or []

        if service_type in stack:
            raise DependencyLoopError(stack + [service_type])

        stack.append(service_type)
        instance = await self._create_service(service_spec, stack)
        stack.pop()

        return instance

    async def _get_dependent_services(
        self, service_spec: ServiceSpec, stack: list
    ) -> dict[str, Any]:
        return {
            name: await self._get(dep, stack) for name, dep in service_spec.dependencies
        }

    async def _create_service(
        self,
        service_spec: ServiceSpec,
        stack: list[type],
    ) -> Any:
        name = service_spec.name

        # check if service exists in cache
        if instance := self._get_from_cache(service_spec.provides, name):
            return instance

        if factory := service_spec.factory:
            dependencies = await self._get_dependent_services(service_spec, stack)

            instance = await maybe_async(factory, **dependencies)
            if inspect.isgenerator(instance):
                generator = instance
                instance = await asyncio.to_thread(next, generator)
                self._setup_generator_finalizer(generator)
            elif inspect.isasyncgen(instance):
                generator = instance
                instance = await anext(generator)
                self._setup_asyncgen_finalizer(generator)

            self.store[service_spec.provides, name] = instance
        else:
            instance = service_spec.service()
            self.store[service_spec.provides, name] = instance

            dependencies = await self._get_dependent_services(service_spec, stack)

            for name, dep_service in dependencies.items():
                setattr(instance, name, dep_service)

            if initializer := service_spec.initializer:
                await maybe_async(initializer, instance)

            self._setup_finalizer(service_spec, instance)

        if service_spec.provides is not Interceptor:
            await self._run_interceptors(instance, service_spec.provides)

        return instance

    def _setup_finalizer(self, service_spec: ServiceSpec, instance: Any):
        if finalizer := service_spec.finalizer:
            self.finalizers.append(maybe_async(finalizer, instance))

    def _setup_generator_finalizer(self, gen: Generator):
        self.finalizers.append(asyncio.to_thread(next, gen, None))

    def _setup_asyncgen_finalizer(self, gen: AsyncGenerator):
        self.finalizers.append(anext(gen, None))

    async def _run_interceptors(self, instance: Any, service_type: type):
        for cls in self.interceptors:
            interceptor = await self.get(
                Interceptor, name=f"{cls.__module__}.{cls.__qualname__}"
            )
            await maybe_async(interceptor.intercept, instance, service_type)

    async def _run_startup(self):
        for service, name in self.startup:
            await self.get(service, name=name)

    async def _run_finalizers(self):
        for finalizer in reversed(self.finalizers):
            await finalizer

        self.finalizers.clear()



from decimal import Decimal
from pathlib import PurePath

from selva.web.converter.decorator import register_param_converter
from selva.web.exception import HTTPBadRequestException


@register_param_converter(str)
class StrParamConverter:
    @staticmethod
    def from_str(value: str) -> str:
        return value

    @staticmethod
    def into_str(data: str) -> str:
        return data


@register_param_converter(int)
class IntParamConverter:
    @staticmethod
    def from_str(value: str) -> int:
        try:
            return int(value)
        except ValueError as err:
            raise HTTPBadRequestException() from err


@register_param_converter(float)
class FloatParamConverter:
    @staticmethod
    def from_str(value: str) -> float:
        try:
            return float(value)
        except ValueError as err:
            raise HTTPBadRequestException() from err


@register_param_converter(Decimal)
class DecimalParamConverter:
    @staticmethod
    def from_str(value: str) -> Decimal:
        try:
            return Decimal(value)
        except ValueError as err:
            raise HTTPBadRequestException() from err


@register_param_converter(bool)
class BoolParamConverter:
    TRUE_VALUES = ["1", "true"]
    FALSE_VALUES = ["0", "false"]

    @staticmethod
    def from_str(value: str) -> bool:
        if value.lower() in BoolParamConverter.TRUE_VALUES:
            return True
        if value.lower() in BoolParamConverter.FALSE_VALUES:
            return False

        raise HTTPBadRequestException()

    @staticmethod
    def into_str(data: bool) -> str:
        return "true" if data else "false"


@register_param_converter(PurePath)
class PurePathParamConverter:
    @staticmethod
    def from_str(value: str) -> PurePath:
        return PurePath(value)

    @staticmethod
    def into_str(data: PurePath) -> str:
        return "/".join(data.parts)



from http import HTTPStatus
from pathlib import Path
from typing import Annotated

from asgikit.responses import Response, respond_stream, respond_text
from jinja2 import Environment, FileSystemLoader, select_autoescape

from selva.configuration import Settings
from selva.di import Inject, service
from selva.ext.templates.jinja.settings import JinjaTemplateSettings
from selva.web.templates import Template


@service(provides=Template)
class JinjaTemplate(Template):
    settings: Annotated[Settings, Inject]
    environment: Environment

    def initialize(self):
        jinja_settings = JinjaTemplateSettings.model_validate(
            self.settings.templates.jinja
        )

        kwargs = jinja_settings.model_dump(exclude_unset=True)

        if "loader" not in kwargs:
            templates_path = [Path(p).absolute() for p in self.settings.templates.paths]
            kwargs["loader"] = FileSystemLoader(templates_path)

        if "autoescape" not in kwargs:
            kwargs["autoescape"] = select_autoescape()

        self.environment = Environment(enable_async=True, **kwargs)

    async def respond(
        self,
        response: Response,
        template_name: str,
        context: dict,
        *,
        status: HTTPStatus = HTTPStatus.OK,
        content_type: str = None,
        stream: bool = False,
    ):
        if content_type:
            response.content_type = content_type
        elif not response.content_type:
            response.content_type = "text/html"

        template = self.environment.get_template(template_name)

        if stream:
            render_stream = template.generate_async(context)
            await respond_stream(response, render_stream, status=status)
        else:
            rendered = await template.render_async(context)
            await respond_text(response, rendered, status=status)

    async def render(self, template_name: str, context: dict) -> str:
        template = self.environment.get_template(template_name)
        return await template.render_async(context)

    async def render_str(self, source: str, context: dict) -> str:
        template = self.environment.from_string(source)
        return await template.render_async(context)



import inspect
from collections.abc import Callable
from typing import Any

from selva.di.service.model import InjectableType


def _type_name(service):
    if (
        inspect.isclass(service)
        or inspect.isfunction(service)
        or inspect.ismethod(service)
    ):
        return service.__qualname__

    return str(service)


class DependencyInjectionError(Exception):
    pass


class DependencyLoopError(DependencyInjectionError):
    def __init__(self, dependency_stack: list[type]):
        loop = " -> ".join([dep.__name__ for dep in dependency_stack])
        super().__init__(f"dependency loop detected: {loop}")


class ServiceNotFoundError(DependencyInjectionError):
    def __init__(
        self,
        service: type,
        name: str = None,
    ):
        message = f"unable to find service '{_type_name(service)}'"
        if name is not None:
            message += f" with name '{name}'"

        super().__init__(message)


class NonInjectableTypeError(DependencyInjectionError):
    def __init__(self, service: Any):
        super().__init__(f"'{service}' is not injectable")


class FactoryMissingReturnTypeError(DependencyInjectionError):
    def __init__(self, factory: Callable):
        super().__init__(f"factory '{_type_name(factory)}' is missing return type")


class ServiceAlreadyRegisteredError(DependencyInjectionError):
    def __init__(self, service: type, name: str = None):
        message = f"service '{_type_name(service)}' is already registered"
        if name:
            message += f" with name '{name}'"

        super().__init__(message)


class TypeVarInGenericServiceError(DependencyInjectionError):
    def __init__(self, provided: type):
        super().__init__(f"{_type_name(provided)} has generic types")


class InvalidServiceTypeError(DependencyInjectionError):
    def __init__(self, service: Any):
        super().__init__(f"object of type {type(service)} is not a valid service")


class ServiceWithoutDecoratorError(DependencyInjectionError):
    def __init__(self, service: InjectableType):
        super().__init__(
            f"service {service.__module__}.{service.__qualname__}"
            " must be decorated with @service"
        )



from redis.asyncio import Redis
from redis.backoff import (
    AbstractBackoff,
    ConstantBackoff,
    DecorrelatedJitterBackoff,
    EqualJitterBackoff,
    ExponentialBackoff,
    FullJitterBackoff,
    NoBackoff,
)
from redis.retry import Retry

from selva.configuration.settings import Settings
from selva.ext.data.redis.settings import RedisSettings

from .settings import BackoffSchema, RetrySchema


def build_backoff(data: BackoffSchema) -> AbstractBackoff:
    if "no_backoff" in data.model_fields_set:
        return NoBackoff()

    if value := data.constant:
        return ConstantBackoff(**value.model_dump(exclude_unset=True))

    if value := data.exponential:
        return ExponentialBackoff(**value.model_dump(exclude_unset=True))

    if value := data.full_jitter:
        return FullJitterBackoff(**value.model_dump(exclude_unset=True))

    if value := data.equal_jitter:
        return EqualJitterBackoff(**value.model_dump(exclude_unset=True))

    if value := data.decorrelated_jitter:
        return DecorrelatedJitterBackoff(**value.model_dump(exclude_unset=True))

    raise ValueError("No value defined for 'backoff'")


def build_retry(data: RetrySchema):
    kwargs = {
        "backoff": build_backoff(data.backoff),
        "retries": data.retries,
    }

    if supported_errors := data.supported_errors:
        kwargs["supported_errors"] = supported_errors

    return Retry(**kwargs)


def make_service(name: str):
    async def redis_service(
        settings: Settings,
    ) -> Redis:
        redis_settings = RedisSettings.model_validate(dict(settings.data.redis[name]))

        kwargs = redis_settings.model_dump(exclude_unset=True)
        if (options := redis_settings.options) and (retry := options.retry):
            kwargs["retry"] = build_retry(retry)

        if url := kwargs.pop("url", ""):
            redis = Redis.from_url(url, **kwargs)
        else:
            redis = Redis(**kwargs)

        await redis.initialize()
        yield redis
        await redis.aclose()

    return redis_service
